/*
 * Copyright (c) 2002 Sandro Sigala.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef WIN32
#include <sys/stat.h>
#include <sys/types.h>
#endif
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "mainwindow.h"
#include "mapwindow.h"
#include "map.h"
#include "paths.h"
#include "version.h"

#include "fxreptoned.xpm"
#include "fxreptoned_mini.xpm"

FXDEFMAP(MainWindow) MainWindowMap[] = {
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_NEW,       MainWindow::onCmdNew),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_OPEN,      MainWindow::onCmdOpen),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_SAVE,      MainWindow::onCmdSave),
  FXMAPFUNC(SEL_UPDATE,    MainWindow::ID_SAVE,      MainWindow::onUpdSave),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_SAVEAS,    MainWindow::onCmdSaveAs),
  FXMAPFUNC(SEL_UPDATE,    MainWindow::ID_SAVEAS,    MainWindow::onUpdSaveAs),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_SCHECK,    MainWindow::onCmdSCheck),
  FXMAPFUNC(SEL_UPDATE,    MainWindow::ID_SCHECK,    MainWindow::onUpdSCheck),
  FXMAPFUNC(SEL_SIGNAL,    MainWindow::ID_QUIT,      MainWindow::onCmdQuit),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_QUIT,      MainWindow::onCmdQuit),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_MAPS,      MainWindow::onCmdMaps),
  FXMAPFUNC(SEL_UPDATE,    MainWindow::ID_MAPS,      MainWindow::onUpdMaps),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_SPRITES,   MainWindow::onCmdSprites),
  FXMAPFUNC(SEL_UPDATE,    MainWindow::ID_SPRITES,   MainWindow::onUpdSprites),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_ABOUT,     MainWindow::onCmdAbout),
  FXMAPFUNC(SEL_CHANGED,   MainWindow::ID_SECTION,   MainWindow::onCmdInfoChange),
  FXMAPFUNC(SEL_CHANGED,   MainWindow::ID_NAME,      MainWindow::onCmdInfoChange),
  FXMAPFUNC(SEL_CHANGED,   MainWindow::ID_AUTHOR,    MainWindow::onCmdInfoChange),
  FXMAPFUNC(SEL_CHANGED,   MainWindow::ID_COMMENT,   MainWindow::onCmdInfoChange),
};

FXIMPLEMENT(MainWindow, FXMainWindow, MainWindowMap, ARRAYNUMBER(MainWindowMap))

MainWindow::MainWindow()
{
}

MainWindow::MainWindow(FXApp *app)
    : FXMainWindow(app, "FX-Repton Editor", 0, 0, DECOR_ALL, 0, 0, 450, 220)
{
	bigicon = new FXXPMIcon(getApp(), fxreptoned_xpm);
  	smallicon = new FXXPMIcon(getApp(), fxreptoned_mini_xpm);
  	setIcon(bigicon);
    	setMiniIcon(smallicon);

       	FXMenubar *menubar;

        FXToolbarShell *dragshell1 = new FXToolbarShell(this, FRAME_RAISED);
        menubar = new FXMenubar(this, dragshell1, FRAME_RAISED|LAYOUT_SIDE_TOP|LAYOUT_FILL_X);
        new FXToolbarGrip(menubar, menubar, FXMenubar::ID_TOOLBARGRIP);

       	FXStatusbar *statusbar;
	statusbar = new FXStatusbar(this, LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|STATUSBAR_WITH_DRAGCORNER|FRAME_RAISED);

        filemenu = new FXMenuPane(this);
        new FXMenuCommand(filemenu, "&New\t\tCreate a new episode.", 0, this, ID_NEW);
 	new FXMenuCommand(filemenu, "&Open...\t\tOpen a user saved episode.", 0, this, ID_OPEN);
        new FXMenuCommand(filemenu, "&Save\tCtrl-S\tSave current open episode.", 0, this, ID_SAVE);
        new FXMenuCommand(filemenu, "S&ave as...\t\tSave current open episode with another name.", 0, this, ID_SAVEAS);
	new FXMenuSeparator(filemenu);
        new FXMenuCommand(filemenu, "Strict &check...\t\tCheck the minimum requirements of playability.", 0, this, ID_SCHECK);
	new FXMenuSeparator(filemenu);
        new FXMenuCommand(filemenu, "&Quit\tAlt-X\tClose this program.", 0, this, ID_QUIT);
        new FXMenuTitle(menubar, "&File", 0, filemenu);

        windowmenu = new FXMenuPane(this);
       	new FXMenuCommand(windowmenu, "Show maps window", 0, this, ID_MAPS);
       	new FXMenuCommand(windowmenu, "Show sprites window", 0, this, ID_SPRITES);
  	new FXMenuTitle(menubar, "&Window", 0, windowmenu);

	helpmenu = new FXMenuPane(this);
        new FXMenuCommand(helpmenu, "&About...\t\tShow information about this program.", 0, this, ID_ABOUT);
        new FXMenuTitle(menubar, "&?", 0, helpmenu);

	FXVerticalFrame *contents =
        	new FXVerticalFrame(this, LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 10,10,10,10, 10,10);

      	shortname_label = new FXLabel(contents, "Short name:", 0, JUSTIFY_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
	FXHorizontalFrame *h1 =
       		new FXHorizontalFrame(contents, LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0, 5,5);
	new FXLabel(h1, "&Menu section:", 0, JUSTIFY_LEFT|LAYOUT_CENTER_Y, 0,0,0,0, 0,0,0,0);
	pkg_edit = new FXTextField(h1, 30, this, ID_SECTION, TEXTFIELD_NORMAL|LAYOUT_FILL_X|LAYOUT_CENTER_Y);
      	FXHorizontalFrame *h2 =
       		new FXHorizontalFrame(contents, LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0, 5,5);
      	new FXLabel(h2, "&Episode name:", 0, JUSTIFY_LEFT|LAYOUT_CENTER_Y, 0,0,0,0, 0,0,0,0);
	name_edit = new FXTextField(h2, 50, this, ID_NAME, TEXTFIELD_NORMAL|LAYOUT_FILL_X|LAYOUT_CENTER_Y);
      	FXHorizontalFrame *h3 =
       		new FXHorizontalFrame(contents, LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0, 5,5);
   	new FXLabel(h3, "&Author:", 0, JUSTIFY_LEFT|LAYOUT_CENTER_Y, 0,0,0,0, 0,0,0,0);
	author_edit = new FXTextField(h3, 30, this, ID_AUTHOR, TEXTFIELD_NORMAL|LAYOUT_FILL_X|LAYOUT_CENTER_Y);
      	FXHorizontalFrame *h4 =
       		new FXHorizontalFrame(contents, LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0, 5,5);
      	new FXLabel(h4, "&Comments:", 0, JUSTIFY_LEFT|LAYOUT_CENTER_Y, 0,0,0,0, 0,0,0,0);
	cmt_edit = new FXTextField(h4, 50, this, ID_COMMENT, TEXTFIELD_NORMAL|LAYOUT_FILL_X|LAYOUT_CENTER_Y);

	mapw = 0;

	for (int i = 0; i < num_sprites; ++i) {
		sprites[i] = new FXBMPImage(getApp(), 0, IMAGE_KEEP|IMAGE_OWNED|IMAGE_SHMI|IMAGE_SHMP,
				    	    sprite_width, sprite_height);
		sprites[i]->create();
	}

 	updateFields();
}

MainWindow::~MainWindow()
{
	delete filemenu;
	delete windowmenu;
	delete helpmenu;
	if (mapw) delete mapw;

	for (int i = 0; i < num_sprites; ++i)
		delete sprites[i];
}

void MainWindow::create()
{
  	FXMainWindow::create();
}

void MainWindow::updateTitle()
{
	FXString s = "FX-Repton Editor";
 	if (!episode.path.empty()) {
  		s += " [";
		s += episode.path;
		s += "]";
	}
 	setTitle(s);
}

void MainWindow::updateFields()
{
	if (episode.isdataloaded) {
		FXString shname = episode.path.after(PATHSEP, -1);
		if (shname.empty()) shname = "(unspecified)";
		shortname_label->setText("Short name: " + shname);
		pkg_edit->enable();
  		name_edit->enable();
    		author_edit->enable();
      		cmt_edit->enable();
	 	pkg_edit->setText(episode.package);
		name_edit->setText(episode.name);
	 	author_edit->setText(episode.author);
		cmt_edit->setText(episode.comments);
	} else {
  		pkg_edit->disable();
  		name_edit->disable();
    		author_edit->disable();
      		cmt_edit->disable();
	}
}

void MainWindow::loadSprites(const FXString& spritespath)
{
	FXString filename = spritespath + PATHSEPSTRING "sprites.bmp";

	FXBMPImage *img = new FXBMPImage(getApp(), 0);

	// Load it
  	FXFileStream stream;
  	if(!stream.open(filename, FXStreamLoad)) {
    		FXMessageBox::error(this, MBOX_OK, "Error", "Cannot load %s", filename.text());
  		exit(1);
     	}

	img->loadPixels(stream);
	stream.close();
	img->create();

	if (img->getWidth() != 8*sprite_width || img->getHeight() != 6*sprite_height) {
    		FXMessageBox::error(this, MBOX_OK, "Error", "The sprite bitmap %s must be 256x192", filename.text());
  		exit(1);
     	}

	// Load the sprites
	for (int y = 0; y < 6; ++y)
		for (int x = 0; x < 8; ++x) {
			int i = y*8 + x;
			FXDCWindow dc(sprites[i]);
		   	dc.setFunction(BLT_SRC);
			dc.drawImage(img, -x*sprite_width, -y*sprite_height);
   		}

	delete img;
}

void MainWindow::saveSprites(const FXString& spritespath)
{
	FXString filename = spritespath + PATHSEPSTRING "sprites.bmp";

	FXBMPImage *img = new FXBMPImage(getApp(), 0, IMAGE_SHMI|IMAGE_SHMP, sprite_width*8, sprite_height*6);
	img->create();

	// Save the sprites
	FXDCWindow dc(img);
   	dc.setFunction(BLT_SRC);
	for (int y = 0; y < 6; ++y)
		for (int x = 0; x < 8; ++x) {
			int i = y*8 + x;
			dc.drawImage(sprites[i], x*sprite_width, y*sprite_height);
   		}

	// Save it
  	FXFileStream stream;
  	if(!stream.open(filename, FXStreamSave)) {
    		FXMessageBox::error(this, MBOX_OK, "Error", "Cannot open for writing %s", filename.text());
  		exit(1);
     	}
      	img->restore();
	img->savePixels(stream);
	stream.close();

	/* TODO (#1#): Why crashes on this? */
//	delete img;
}

bool MainWindow::strictCheckTransporters(int no, Map *m)
{
	bool err = false;
	FXString s = FXStringFormat("%d: \"%s\"", no, m->password.text());

	for (int y = 0; y < m->height; ++y)
		for (int x = 0; x < m->width; ++x) {
			if (m->xy(x, y) == Sprite_Transporter) {
				int count = 0;
				typedef vector<Transporter *>::const_iterator J;
				for (J j = m->transporters.begin(); j != m->transporters.end(); ++j) {
					Transporter *t = *j;
					if (t->orig_x == x && t->orig_y == y)
						++count;
				}

				if (count < 1) {
					FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has one undefined transporter at %d, %d!", s.text(), x+1, y+1);
					err = true;
				}
				if (count > 1) {
					FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has %d definitions of transporter at %d, %d!", s.text(), count, x+1, y+1);
					err = true;
				}
		}
 	}

	typedef vector<Transporter *>::const_iterator J;
	for (J j = m->transporters.begin(); j != m->transporters.end(); ++j) {
		Transporter *t = *j;
		if (t->orig_x < 0 | t->orig_y < 0 || t->orig_x >= m->width || t->orig_y >= m->height) {
			FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has one transporter with origin outside the map!", s.text());
			err = true;
     		}
		if (m->xy(t->orig_x, t->orig_y) != Sprite_Transporter) {
			FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has one transporter defined at %d, %d but no transporter is found in the map at that location!", s.text(), t->orig_x, t->orig_y);
			err = true;
     		}
		if (t->dest_x < 0 | t->dest_y < 0 || t->dest_x >= m->width || t->dest_y >= m->height) {
			FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has one transporter located at %d, %d with destination outside the map!", s.text(), t->orig_x, t->orig_y);
			err = true;
     		}
	}

	return !err;
}

bool MainWindow::strictCheck(bool showok)
{
	bool err = false;

	if (episode.package.empty()) {
		FXMessageBox::error(this, MBOX_OK, "Error", "The episode section name must be specified!");
		err = true;
 	}

	if (episode.name.empty()) {
		FXMessageBox::error(this, MBOX_OK, "Error", "The episode name must be specified!");
		err = true;
 	}

	typedef vector<Map *>::const_iterator I;
	for (I i = episode.maps.begin(); i != episode.maps.end(); ++i) {
		Map *m = *i;
		FXString s = FXStringFormat("#%d: \"%s\"", i-episode.maps.begin()+1, m->password.text());

		if (m->password.empty()) {
			FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has no password specified!", s.text());
			err = true;
 		}

 		if (m->time_limit <= 0) {
			FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has an invalid time limit!", s.text());
			err = true;
		}

		int num_reptons = 0, num_diamonds = 0;
		int num_spirits = 0, num_cages = 0, num_bombs = 0;
		int num_strongboxes = 0, num_keys = 0;
		for (int y = 0; y < m->height; ++y)
			for (int x = 0; x < m->width; ++x)
				switch (m->xy(x, y)) {
				case Sprite_Repton:
					++num_reptons;
					break;
				case Sprite_Diamond:
					++num_diamonds;
					break;
				case Sprite_Spirit:
					++num_spirits;
					break;
				case Sprite_Cage:
					++num_cages;
					break;
				case Sprite_Bomb:
					++num_bombs;
					break;
				case Sprite_Strongbox:
					++num_strongboxes;
					break;
				case Sprite_Key:
					++num_keys;
					break;
				default:; // Avoid warnings
				}

 		if (num_reptons == 0) {
			FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has no Repton!", s.text());
			err = true;
		}

 		if (num_reptons > 1) {
			FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has more than one Repton!", s.text());
			err = true;
		}

 		if (num_bombs == 0) {
			FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has no end level object!", s.text());
			err = true;
		}

 		if (num_cages > num_spirits) {
			FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has more cages than monsters (spirits)!", s.text());
			err = true;
		}

 		if (num_strongboxes > 0 && num_keys < 1) {
			FXMessageBox::error(this, MBOX_OK, "Error", "The map %s has strongboxes but no keys!", s.text());
			err = true;
		}

		if (!strictCheckTransporters(i-episode.maps.begin()+1, m))
			err = true;
	}

	// Check for duplicated passwords
	for (I i = episode.maps.begin(); i != episode.maps.end(); ++i) {
		Map *m1 = *i;
		for (I j = i; j != episode.maps.end(); ++j) {
			Map *m2 = *j;
			if (m1 == m2)
				continue;
			if (m1->password == m2->password) {
				FXMessageBox::error(this, MBOX_OK, "Error", "The maps #%d and #%d have the same password \"%s\"!",
    						    i-episode.maps.begin()+1, j-episode.maps.begin()+1, m1->password.text());
      				err = true;
			}
		}
	}

	if (showok && !err)
		FXMessageBox::information(this, MBOX_OK, "Strict Check", "No errors found!");

	return !err;
}

bool MainWindow::save(bool saveas)
{
	episode.isbroken = !strictCheck(false);

	if (saveas || episode.path.empty()) {
		FXString shname = episode.path.after(PATHSEP, -1);
		vector<Episode> v;
		Episode::loadEpisodesList(v, FXFile::getHomeDirectory() + PATHSEPSTRING PATH_USER_EPISODES);
  		for (;;) {
    		reask:
			if (!FXInputDialog::getString(shname, this, "FX-Repton Editor", "Enter episode short name (alphanumeric only):"))
 				return false;
			shname = shname.trim();
			if (shname.empty())
				continue;
			for (int i = 0; i < shname.length(); ++i)
				if (!isalnum(shname[i]) && shname[i] != '_' && shname[i] != '-')
					goto reask;
		  	typedef vector<Episode>::const_iterator I;
			for (I i = v.begin(); i != v.end(); ++i) {
				const Episode& e = *i;
				if (e.path.after(PATHSEP, -1) == shname) {
					int r = FXMessageBox::question(this, MBOX_YES_NO_CANCEL, "FX-Repton Editor", "An episode named %s already exists. Do you want to overwrite it?", shname.text());
					if (r == MBOX_CLICKED_CANCEL)
						return false;
					else if (r == MBOX_CLICKED_YES)
						goto nameok;
					else
   						goto reask;
 				}
			}
			break;
   		}

	nameok:
		episode.path = FXFile::getHomeDirectory() + PATHSEPSTRING PATH_USER_EPISODES PATHSEPSTRING + shname;
	}

#ifdef WIN32
	mkdir((FXFile::getHomeDirectory() + PATHSEPSTRING PATH_USER_EPISODES).text());
	mkdir(episode.path.text());
#else
	mkdir((FXFile::getHomeDirectory() + PATHSEPSTRING PATH_USER_EPISODES).text(), 0755);
	mkdir(episode.path.text(), 0755);
#endif
	episode.save(episode.path);
	saveSprites(episode.path);
	updateTitle();
	updateFields();

	return true;
}

bool MainWindow::saveIfChanged()
{
	if (episode.isModified()) {
		int r = FXMessageBox::question(this, MBOX_YES_NO_CANCEL, "FX-Repton Editor", "The episode has been changed. Do you want to save it?");
		if (r == MBOX_CLICKED_CANCEL)
			return false;
		else if (r == MBOX_CLICKED_YES)
   			return save(false);
  	}
	return true;
}

FXString MainWindow::chooseEpisode(const vector<Episode>& v, bool templ, bool *incmaps)
{
	FXDialogBox choose(this, (templ ? "Choose Template Episode" : "Open Episode"), DECOR_TITLE|DECOR_BORDER,
 		0,0,300,300, 0,0,0,0, 0,0);

  	FXVerticalFrame *contents =
        	new FXVerticalFrame(&choose, LAYOUT_SIDE_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 10,10,10,10, 0,0);
    	new FXLabel(contents, (templ ? "Choose episode as template:" : "Choose episode:"), 0, JUSTIFY_LEFT|LAYOUT_FILL_X);
  	FXHorizontalFrame *epframe =
   		new FXHorizontalFrame(contents, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_FILL_X,
   			0,0,0,0, 0,0,0,0, 0,0);
     	FXList *eplist = new FXList(epframe, 10, 0, 0, LIST_BROWSESELECT|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  	typedef vector<Episode>::const_iterator I;
	for (I i = v.begin(); i != v.end(); ++i) {
		const Episode& e = *i;
		FXString package = e.package;
		FXString name = e.name;
  		if (package.findf(']') != -1)
			package = package.after(']').trim();
  		if (name.findf(']') != -1)
			name = name.after(']').trim();
		FXString s = FXStringFormat("%s: %s (%s)", package.text(), name.text(), e.path.after(PATHSEP, -1).text());
		eplist->appendItem(s);
	}
	FXCheckButton *maps_check = 0;
	if (templ) {
		maps_check = new FXCheckButton(contents, "&Include maps from template", 0, 0, CHECKBUTTON_NORMAL, 0,0,0,0, 0,0,10,5);
		maps_check->setCheck(FALSE);
	}
	FXHorizontalFrame *buttons =
        	new FXHorizontalFrame(contents, LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,10,0, 10,10);
        new FXButton(buttons, "Cancel", 0, &choose,
	        FXDialogBox::ID_CANCEL, BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,
              	0,0,0,0, 32,32,5,5);
	FXButton *ok = new FXButton(buttons, "OK", 0, &choose,
	        FXDialogBox::ID_ACCEPT, BUTTON_INITIAL|BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,
         	0,0,0,0, 32,32,5,5);
        ok->setFocus();
        if (choose.execute(PLACEMENT_OWNER)) {
	 	if (templ)
 			*incmaps = (maps_check->getCheck() == TRUE);
	       	return v[eplist->getCurrentItem()].path;
 	}
 	return FXString::null;
}

long MainWindow::onCmdNew(FXObject *, FXSelector, void *)
{
	if (!saveIfChanged())
		return 1;

	vector<Episode> v;
	Episode::loadEpisodesList(v, PATH_EPISODES);
	Episode::loadEpisodesList(v, FXFile::getHomeDirectory() + PATHSEPSTRING PATH_USER_EPISODES);

	if (v.empty())
		FXMessageBox::information(this, MBOX_OK, "FX-Repton Editor", "No episodes found! Have you installed FX-Repton correctly?");
	else {
		bool incmaps;
		FXString epath = chooseEpisode(v, true, &incmaps);
		if (!epath.empty()) {
			if (incmaps) {
				episode.load(epath.text());
    				episode.author = episode.comments = episode.path = FXString::null;
			} else {
				episode.clear();
				Map *m = new Map();
				m->newDummyMap(28, 24);
				episode.appendMap(m);
				episode.isdataloaded = true;
			}
			episode.package = "Other";
			episode.name = "Untitled";
			loadSprites(epath);
			updateTitle();
			updateFields();
			if (mapw) {
				mapw->selectMap(0);
				mapw->spritesChanged();
			}
			pkg_edit->setFocus();
		}
	}

	return 1;
}

long MainWindow::onCmdOpen(FXObject *, FXSelector, void *)
{
	FXint x, y;
 	FXuint buttons;
	getCursorPosition(x, y, buttons);

	if (!saveIfChanged())
		return 1;

	vector<Episode> v;
	if (buttons & SHIFTMASK) // Secret! :-)
  		Episode::loadEpisodesList(v, PATH_EPISODES);
	Episode::loadEpisodesList(v, FXFile::getHomeDirectory() + PATHSEPSTRING PATH_USER_EPISODES);

	if (v.empty())
		FXMessageBox::information(this, MBOX_OK, "FX-Repton Editor", "No user episodes found! Create one with the \"New\" command!");
	else {
		FXString epath = chooseEpisode(v);
		if (!epath.empty()) {
			episode.load(epath.text());
			loadSprites(epath);
			updateTitle();
			updateFields();
			if (mapw) {
  				mapw->selectMap(0);
				mapw->spritesChanged();
			}
			pkg_edit->setFocus();
		}
 	}
	return 1;
}

long MainWindow::onCmdSave(FXObject *, FXSelector, void *)
{
	save(false);
	return 1;
}

long MainWindow::onUpdSave(FXObject *sender, FXSelector, void *)
{
  	FXuint enabled = (episode.isdataloaded && episode.isModified()) ? ID_ENABLE : ID_DISABLE;
   	sender->handle(this, MKUINT(enabled, SEL_COMMAND), 0);
	return 1;
}

long MainWindow::onCmdSaveAs(FXObject *, FXSelector, void *)
{
	save(true);
	return 1;
}

long MainWindow::onUpdSaveAs(FXObject *sender, FXSelector, void *)
{
  	FXuint enabled = episode.isdataloaded ? ID_ENABLE : ID_DISABLE;
   	sender->handle(this, MKUINT(enabled, SEL_COMMAND), 0);
	return 1;
}

long MainWindow::onCmdSCheck(FXObject *, FXSelector, void *)
{
	strictCheck(true);
	return 1;
}

long MainWindow::onUpdSCheck(FXObject *sender, FXSelector, void *)
{
  	FXuint enabled = episode.isdataloaded ? ID_ENABLE : ID_DISABLE;
   	sender->handle(this, MKUINT(enabled, SEL_COMMAND), 0);
	return 1;
}

long MainWindow::onCmdQuit(FXObject *, FXSelector, void *)
{
	if (saveIfChanged())
		getApp()->exit(0);
 	return 1;
}

long MainWindow::onCmdMaps(FXObject *, FXSelector, void *)
{
	if (!mapw) {
		mapw = new MapWindow(getApp(), "FX-Repton Editor - Map Editor", &episode, sprites, bigicon, smallicon);
		mapw->create();
	}
	if (!mapw->shown())
      		mapw->show(PLACEMENT_OWNER);
	else
      		mapw->hide();
	return 1;
}

long MainWindow::onUpdMaps(FXObject *sender, FXSelector, void *)
{
	FXint checked = mapw && mapw->shown();
	sender->handle(this, MKUINT(ID_SETINTVALUE, SEL_COMMAND), (void*)&checked);
  	FXuint enabled = episode.isdataloaded ? ID_ENABLE : ID_DISABLE;
   	sender->handle(this, MKUINT(enabled, SEL_COMMAND), 0);
	return 1;
}

long MainWindow::onCmdSprites(FXObject *, FXSelector, void *)
{
	return 1;
}

long MainWindow::onUpdSprites(FXObject *sender, FXSelector, void *)
{
   	sender->handle(this, MKUINT(ID_DISABLE, SEL_COMMAND), 0);
	return 1;
}

long MainWindow::onCmdAbout(FXObject *, FXSelector, void *)
{
	FXDialogBox about(this, "About FX-Repton Editor", DECOR_TITLE|DECOR_BORDER,
 		0,0,0,0, 0,0,0,0, 0,0);

  	FXVerticalFrame *info2 =
        	new FXVerticalFrame(&about, LAYOUT_SIDE_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 10,10,10,10, 0,0);
  	FXHorizontalFrame *cont =
        	new FXHorizontalFrame(info2, LAYOUT_SIDE_RIGHT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 0,0);
  	FXHorizontalFrame *infoandicon =
        	new FXHorizontalFrame(cont, LAYOUT_SIDE_RIGHT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 0,0);
	new FXLabel(infoandicon, FXString::null, bigicon, LAYOUT_FILL_X,
		0,0,0,0, 0,0,0,0);
  	FXVerticalFrame *info =
   		new FXVerticalFrame(infoandicon, LAYOUT_SIDE_RIGHT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
   			0,0,0,0, 10,10,10,10, 0,0);
    	new FXLabel(info, "FX-Repton Editor " VERSION "\n", 0, JUSTIFY_LEFT|LAYOUT_FILL_X);
      	new FXHorizontalSeparator(info, SEPARATOR_GROOVE|LAYOUT_FILL_X);
        new FXLabel(info,
        	"\nCopyright "
#ifdef WIN32
        	""
#else
         "(C)"
#endif
         " 2002 Sandro Sigala <sandro@sigala.it>.\n\n"
        	"Latest version can be found at http://www.sigala.it/sandro.",
      		0, JUSTIFY_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y);
        FXButton *button = new FXButton(cont, "OK", 0, &about,
	        FXDialogBox::ID_ACCEPT, BUTTON_INITIAL|BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,
         	0,0,0,0, 32,32,5,5);

  	FILE *fp;
      	if ((fp = fopen(PATH_DATA_DIR PATHSEPSTRING "Copying", "r")) != 0) {
  	      	FXuint size, n;
  	      	FXchar *text;
  	      	fseek(fp, 0, SEEK_END);
    	      	size = ftell(fp);
  	      	fseek(fp, 0, SEEK_SET);
      	      	FXCALLOC(&text,FXchar,size+1);
      	      	n = fread(text, 1, size, fp);
      	      	fclose(fp);
        	new FXLabel(info2, "License:", 0, LAYOUT_SIDE_TOP|JUSTIFY_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y);
        	FXVerticalFrame *textf =
        		new FXVerticalFrame(info2, FRAME_SUNKEN|LAYOUT_SIDE_TOP|LAYOUT_FIX_HEIGHT|LAYOUT_FILL_X,
       	 			0,0,0,200, 0,0,0,0, 0,0);
                FXText *textv = new FXText(textf, 0, 0, LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y);
      	      	textv->setText(text, n);
                textv->setEditable(false);
      	      	FXFREE(&text);
        }

        new FXLabel(info2,
        	"\nThis software uses the FOX Toolkit Library (http://www.fox-toolkit.org).",
      		0, JUSTIFY_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y);

        button->setFocus();
      	about.execute(PLACEMENT_OWNER);

	return 1;
}

long MainWindow::onCmdInfoChange(FXObject *, FXSelector, void *)
{
	episode.package = pkg_edit->getText().trim();
	episode.name = name_edit->getText().trim();
	episode.author = author_edit->getText().trim();
	episode.comments = cmt_edit->getText().trim();
	episode.ismodified = true;
	return 1;
}

