/*
 * Copyright (c) 2002 Sandro Sigala.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapwindow.h"
#include "mapviewer.h"
#include "map.h"

FXDEFMAP(MapWindow) MapWindowMap[] = {
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE, MapWindow::ID_TOOLS,   MapWindow::onToolsMouseUp),
  FXMAPFUNC(SEL_CHANGED,           MapWindow::ID_MAP,     MapWindow::onMapChange),
  FXMAPFUNC(SEL_MOTION,	           MapWindow::ID_MAP,     MapWindow::onMapMouseMove),
  FXMAPFUNC(SEL_COMMAND,           MapWindow::ID_MAPSLIST,MapWindow::onMapListChange),
  FXMAPFUNC(SEL_COMMAND,           MapWindow::ID_MAPINS,  MapWindow::onCmdMapInsert),
  FXMAPFUNC(SEL_COMMAND,           MapWindow::ID_MAPDUP,  MapWindow::onCmdMapDuplicate),
  FXMAPFUNC(SEL_COMMAND,           MapWindow::ID_MAPRSZ,  MapWindow::onCmdMapResize),
  FXMAPFUNC(SEL_COMMAND,           MapWindow::ID_MAPDEL,  MapWindow::onCmdMapDelete),
  FXMAPFUNC(SEL_COMMAND,           MapWindow::ID_MAPUP,   MapWindow::onCmdMapUp),
  FXMAPFUNC(SEL_COMMAND,           MapWindow::ID_MAPDOWN, MapWindow::onCmdMapDown),
  FXMAPFUNC(SEL_COMMAND,           MapWindow::ID_TRANSCHG,MapWindow::onCmdTransChange),
  FXMAPFUNC(SEL_COMMAND,           MapWindow::ID_TRANSDEL,MapWindow::onCmdTransDelete),
  FXMAPFUNC(SEL_COMMAND,           MapWindow::ID_READONLY,MapWindow::onCmdInfoChange),
  FXMAPFUNC(SEL_CHANGED,           MapWindow::ID_PWD,     MapWindow::onCmdInfoChange),
  FXMAPFUNC(SEL_CHANGED,           MapWindow::ID_COMMENT, MapWindow::onCmdInfoChange),
  FXMAPFUNC(SEL_CHANGED,           MapWindow::ID_TIME,    MapWindow::onCmdInfoChange),
};

FXIMPLEMENT(MapWindow, FXTopWindow, MapWindowMap, ARRAYNUMBER(MapWindowMap))

MapWindow::MapWindow()
{
}

MapWindow::MapWindow(FXApp *a, const FXString& name, Episode *e, FXBMPImage **s, FXIcon *icon, FXIcon *miniicon)
	: FXTopWindow(a, name, icon, miniicon, DECOR_ALL, 0,0,700,600, 0,0,0,0, 0,0)
{
	episode = e;
	sprites = s;

	tools = new FXBMPImage(getApp(), 0, IMAGE_KEEP|IMAGE_OWNED|IMAGE_SHMI|IMAGE_SHMP,
		    	    sprite_width+10, (sprite_height+5)*num_map_sprites + 5);
   	tools->create();

       	FXStatusbar *statusbar;
	statusbar = new FXStatusbar(this, LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|STATUSBAR_WITH_DRAGCORNER|FRAME_RAISED);
  	pos_label = new FXLabel(statusbar, "", 0, FRAME_SUNKEN|LAYOUT_RIGHT|LAYOUT_CENTER_Y|LAYOUT_FIX_WIDTH, 0,0,50,0);

	FXVerticalFrame *contents =
        	new FXVerticalFrame(this, LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 10,10,10,10, 10,10);
  	FXHorizontalFrame *contentsh =
        	new FXHorizontalFrame(contents, LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 5,5);
	FXVerticalFrame *toolsvf =
        	new FXVerticalFrame(contentsh, LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 5,5);
//      	new FXLabel(toolsvf, "Tools", 0, JUSTIFY_LEFT|LAYOUT_FILL_X);
	FXHorizontalFrame *toolsframe =
   		new FXHorizontalFrame(toolsvf, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_Y,
   			0,0,0,0, 0,0,0,0, 0,0);
	toolsview = new FXImageView(toolsframe, tools, this, ID_TOOLS, LAYOUT_FIX_WIDTH|LAYOUT_FILL_Y,
				  0,0,sprite_width+10+16,0);
	toolsview->setBackColor(FXRGB(0, 0, 0));
	toolsview->verticalScrollbar()->setLine(sprite_height + 5);

  	FXHorizontalFrame *mapframe =
   		new FXHorizontalFrame(contentsh, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_Y|LAYOUT_FILL_X,
   			0,0,0,0, 0,0,0,0, 0,0);
	mapview = new MapViewer(mapframe, this, ID_MAP, LAYOUT_FILL_X|LAYOUT_FILL_Y,
			0,0,0,0);
	mapview->setBackColor(FXRGB(0, 0, 0));

  	FXHorizontalFrame *contentsh2 =
        	new FXHorizontalFrame(contents, LAYOUT_FILL_X|LAYOUT_FIX_HEIGHT,
       	 		0,0,0,160, 0,0,0,0, 10,10);
	FXVerticalFrame *contentsh21 =
        	new FXVerticalFrame(contentsh2, LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 0,0);
	FXVerticalFrame *contentsh22 =
        	new FXVerticalFrame(contentsh2, LAYOUT_FILL_Y|PACK_UNIFORM_WIDTH,
       	 		0,0,0,0, 0,0,0,0, 5,5);
      	new FXVerticalSeparator(contentsh2, SEPARATOR_GROOVE|LAYOUT_FILL_Y);
	FXVerticalFrame *contentsh23 =
        	new FXVerticalFrame(contentsh2, LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 0,0);
	FXVerticalFrame *contentsh24 =
        	new FXVerticalFrame(contentsh2, LAYOUT_FILL_Y|PACK_UNIFORM_WIDTH,
       	 		0,0,0,0, 0,0,0,0, 5,5);
      	new FXVerticalSeparator(contentsh2, SEPARATOR_GROOVE|LAYOUT_FILL_Y);
	FXVerticalFrame *contentsh25 =
        	new FXVerticalFrame(contentsh2, LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 0,0);

      	new FXLabel(contentsh21, "Maps:", 0, JUSTIFY_LEFT|LAYOUT_FILL_X);
	FXHorizontalFrame *mapsframe =
   		new FXHorizontalFrame(contentsh21, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y,
   			0,0,0,0, 0,0,0,0, 0,0);
   	maps_list = new FXList(mapsframe, 20, this, ID_MAPSLIST, LIST_BROWSESELECT|LAYOUT_FILL_X|LAYOUT_FILL_Y);

	new FXHorizontalFrame(contentsh22, LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0, 0,0);
        map_ins = new FXButton(contentsh22, "&New\t\tInsert a new map.", 0, this, ID_MAPINS, FRAME_RAISED|FRAME_THICK);
        map_dup = new FXButton(contentsh22, "&Duplicate\t\tDuplicate the selected map.", 0, this, ID_MAPDUP, FRAME_RAISED|FRAME_THICK);
        map_rsz = new FXButton(contentsh22, "&Resize\t\tResize the selected map.", 0, this, ID_MAPRSZ, FRAME_RAISED|FRAME_THICK);
        map_del = new FXButton(contentsh22, "D&elete\t\tDelete the selected map.", 0, this, ID_MAPDEL, FRAME_RAISED|FRAME_THICK);
        map_up = new FXButton(contentsh22, "Move &up\t\tMove the selected map up one position.", 0, this, ID_MAPUP, FRAME_RAISED|FRAME_THICK);
        map_down = new FXButton(contentsh22, "Move d&own\t\tMove the selected map down one position.", 0, this, ID_MAPDOWN, FRAME_RAISED|FRAME_THICK);
	new FXHorizontalFrame(contentsh22, LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0, 0,0);

      	new FXLabel(contentsh23, "Transporters:", 0, JUSTIFY_LEFT|LAYOUT_FILL_X);
  	FXHorizontalFrame *transframe =
   		new FXHorizontalFrame(contentsh23, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y,
   			0,0,0,0, 0,0,0,0, 0,0);
   	trans_list = new FXList(transframe, 20, this, ID_TRANSLIST, LIST_BROWSESELECT|LAYOUT_FILL_X|LAYOUT_FILL_Y);

	new FXHorizontalFrame(contentsh24, LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0, 0,0);
        trans_chg = new FXButton(contentsh24, "&Change\t\tChange the transporter destionation coordinates.", 0, this, ID_TRANSCHG, FRAME_RAISED|FRAME_THICK);
        trans_del = new FXButton(contentsh24, "De&lete\t\tDelete the selected transporter.", 0, this, ID_TRANSDEL, FRAME_RAISED|FRAME_THICK);
	new FXHorizontalFrame(contentsh24, LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0, 0,0);

      	mapsize_label = new FXLabel(contentsh25, "Map size: ", 0, JUSTIFY_LEFT|LAYOUT_FILL_X, 0,0,0,0, 0,0,0,0);
	new FXHorizontalFrame(contentsh25, LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0, 0,0);

   	readonly_check = new FXCheckButton(contentsh25, "Re&ad only", this, ID_READONLY, CHECKBUTTON_NORMAL);
      	FXHorizontalFrame *h1 =
       		new FXHorizontalFrame(contentsh25, LAYOUT_FILL_X, 0,0,0,0, 0,0,5,5, 5,5);
   	new FXLabel(h1, "Password:", 0, JUSTIFY_LEFT|LAYOUT_CENTER_Y, 0,0,0,0, 0,0,0,0);
	pwd_edit = new FXTextField(h1, 20, this, ID_PWD, TEXTFIELD_NORMAL|LAYOUT_FILL_X);
      	FXHorizontalFrame *h2 =
       		new FXHorizontalFrame(contentsh25, LAYOUT_FILL_X, 0,0,0,0, 0,0,5,5, 5,5);
      	new FXLabel(h2, "Time limit (seconds):", 0, JUSTIFY_LEFT|LAYOUT_CENTER_Y, 0,0,0,0, 0,0,0,0);
	time_edit = new FXTextField(h2, 4, this, ID_TIME, TEXTFIELD_NORMAL|TEXTFIELD_INTEGER|TEXTFIELD_LIMITED);
      	new FXLabel(contentsh25, "Comments:", 0, JUSTIFY_LEFT, 0,0,0,0, 0,0,0,0);
	cmt_edit = new FXTextField(contentsh25, 30, this, ID_COMMENT, TEXTFIELD_NORMAL|LAYOUT_FILL_X);

	currenttool = Sprite_Rock;
	redrawTools();
	selectMap(0);
}

MapWindow::~MapWindow()
{
 	delete tools;
}

void MapWindow::create()
{
	FXTopWindow::create();
}

void MapWindow::redrawTools()
{
	FXDCWindow dc(tools);
	dc.setForeground(FXRGB(0, 0, 0));
	dc.fillRectangle(0, 0, tools->getWidth(), tools->getHeight());
   	dc.setFunction(BLT_SRC);
  	for (int i = 0; i < num_map_sprites; ++i)
		dc.drawImage(sprites[i], 5, i*(sprite_height+5) + 5);

	dc.setForeground(FXRGB(255, 255, 255));
	dc.drawRectangle(3, currenttool*(sprite_height + 5) + 5 - 2, sprite_width + 4, sprite_height + 4);

	toolsview->update();
}

void MapWindow::redrawMap()
{
	mapview->update();
}

void MapWindow::updateButtons()
{
	Map *m = episode->maps[currentmap];
	if (m->readonly || maps_list->getNumItems() < 2)
  		map_del->disable();
    	else
  		map_del->enable();
	if (maps_list->getCurrentItem() > 0)
  		map_up->enable();
	else
  		map_up->disable();
	if (maps_list->getCurrentItem() < maps_list->getNumItems() - 1)
  		map_down->enable();
	else
  		map_down->disable();
	if (m->readonly || trans_list->getNumItems() < 1) {
  		trans_chg->disable();
  		trans_del->disable();
	} else {
		trans_chg->enable();
  		trans_del->enable();
	}
	if (m->readonly) {
		map_rsz->disable();
 	} else {
		map_rsz->enable();
  	}
  	/* TODO (#1#): when finished... */
  	map_rsz->disable();
}

long MapWindow::onMapChange(FXObject *, FXSelector, void *)
{
	updateTransList();
	updateButtons();
	return 1;
}

long MapWindow::onMapMouseMove(FXObject *, FXSelector, void *ptr)
{
  	FXEvent *ev = (FXEvent *)ptr;
	int x = (ev->win_x + mapview->horizontalScrollbar()->getPosition()) / sprite_width;
	int y = (ev->win_y + mapview->verticalScrollbar()->getPosition()) / sprite_height;
	Map *m = episode->maps[currentmap];
	if (x >= 0 && y >= 0 && x < m->width && y < m->height) {
		pos_label->setText(FXStringFormat("%d, %d", x+1, y+1));
	}
	return 1;
}

long MapWindow::onToolsMouseUp(FXObject *, FXSelector, void *ptr)
{
  	FXEvent *ev = (FXEvent *)ptr;
	if (ev->win_x > 2 && ev->win_x < sprite_width + 4) {
		int y = ev->win_y + toolsview->verticalScrollbar()->getPosition();
		currenttool = (Sprite)(y / (sprite_height + 5));
  		currenttool = FXMIN(currenttool, Sprite_Spirit);
  		mapview->setTool(currenttool);
		redrawTools();
    	}
	return 1;
}

long MapWindow::onMapListChange(FXObject *, FXSelector, void *)
{
	if (currentmap != maps_list->getCurrentItem())
		selectMap(maps_list->getCurrentItem());
	return 1;
}

long MapWindow::onCmdMapInsert(FXObject *, FXSelector, void *)
{
	FXDialogBox dlg(this, "Enter Map Geometry", DECOR_TITLE|DECOR_BORDER,
 		0,0,0,0, 0,0,0,0, 0,0);

  	FXVerticalFrame *contents =
        	new FXVerticalFrame(&dlg, LAYOUT_SIDE_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 10,10,10,10, 10,10);

 	FXGroupBox *group = new FXGroupBox(contents, "Geometry", GROUPBOX_NORMAL|FRAME_GROOVE,
  			0,0,0,0, 5,5,5,5, 10,10);

  	FXVerticalFrame *gcontents =
        	new FXVerticalFrame(group, LAYOUT_SIDE_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 0,0);

    	FXHorizontalFrame *h1 =
        	new FXHorizontalFrame(gcontents, LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 10,10);

	new FXLabel(h1, "Width:", 0, JUSTIFY_LEFT|LAYOUT_FILL_X);
	FXSpinner *wspin = new FXSpinner(h1, 3, 0, 0, SPIN_NORMAL|FRAME_SUNKEN|FRAME_THICK);
	wspin->setRange(2, 999);
    	wspin->setValue(episode->maps[currentmap]->width);

	new FXLabel(h1, "Height:", 0, JUSTIFY_LEFT|LAYOUT_FILL_X);
	FXSpinner *hspin = new FXSpinner(h1, 3, 0, 0, SPIN_NORMAL|FRAME_SUNKEN|FRAME_THICK);
	hspin->setRange(2, 999);
	hspin->setValue(episode->maps[currentmap]->height);

	FXCheckButton *min_check;
	min_check = new FXCheckButton(contents, "&Include minimal objects", 0, 0, CHECKBUTTON_NORMAL, 0,0,0,0, 0,0,5,5);
	min_check->setCheck(TRUE);

	FXHorizontalFrame *buttons =
        	new FXHorizontalFrame(contents, LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 10,10);
        new FXButton(buttons, "Cancel", 0, &dlg,
	        FXDialogBox::ID_CANCEL, BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,
              	0,0,0,0, 32,32,5,5);
	FXButton *ok = new FXButton(buttons, "OK", 0, &dlg,
	        FXDialogBox::ID_ACCEPT, BUTTON_INITIAL|BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,
         	0,0,0,0, 32,32,5,5);
        ok->setFocus();
        if (dlg.execute(PLACEMENT_OWNER)) {
        	Map *m = new Map();
		if (min_check->getCheck() == TRUE)
  			m->newDummyMap(wspin->getValue(), hspin->getValue());
		else
  			m->newMap(wspin->getValue(), hspin->getValue());
		m->ismodified = true;
    		episode->appendMap(m);
      		selectMap(episode->maps.size() - 1);
   	}

	return 1;
}

long MapWindow::onCmdMapDuplicate(FXObject *, FXSelector, void *)
{
    	Map *m = episode->maps[currentmap];
    	Map *m2 = new Map;
    	m2->copy(m);
    	m2->password = m->password + "COPY";
	m2->ismodified = true;
	episode->maps.insert(episode->maps.begin() + maps_list->getCurrentItem() + 1, m2);
	selectMap(currentmap + 1);
	return 1;
}

long MapWindow::onCmdMapResize(FXObject *, FXSelector, void *)
{
	/* TODO (#1#): MapResize */
    	Map *m = episode->maps[currentmap];
	m->resize(m->width - 1, m->height - 1, true, true);
	m->ismodified = true;
	return 1;
}

long MapWindow::onCmdMapDelete(FXObject *, FXSelector, void *)
{
    	Map *m = episode->maps[currentmap];
	if (FXMessageBox::question(this, MBOX_YES_NO, "FX-Repton Editor",
 	    "Are you sure to delete the #%d \"%s\" map?", currentmap+1, m->password.text()) == MBOX_CLICKED_YES) {
 		episode->maps.erase(episode->maps.begin() + maps_list->getCurrentItem());
		episode->ismodified = true;
		delete m;
		if (maps_list->getCurrentItem() == maps_list->getNumItems() - 1)
			selectMap(maps_list->getNumItems() - 2);
		else
			selectMap(maps_list->getCurrentItem());
 	}
	return 1;
}

long MapWindow::onCmdMapUp(FXObject *, FXSelector, void *)
{
    	Map *m = episode->maps[currentmap];
	episode->maps.erase(episode->maps.begin() + maps_list->getCurrentItem());
	episode->maps.insert(episode->maps.begin() + maps_list->getCurrentItem() - 1, m);
	episode->ismodified = true;
	selectMap(currentmap - 1);
	return 1;
}

long MapWindow::onCmdMapDown(FXObject *, FXSelector, void *)
{
    	Map *m = episode->maps[currentmap];
	episode->maps.erase(episode->maps.begin() + maps_list->getCurrentItem());
	episode->maps.insert(episode->maps.begin() + maps_list->getCurrentItem() + 1, m);
	episode->ismodified = true;
	selectMap(currentmap + 1);
	return 1;
}

long MapWindow::onCmdTransChange(FXObject *, FXSelector, void *)
{
	Transporter *t;
	Map *m = episode->maps[currentmap];
	t = m->transporters[trans_list->getCurrentItem()];
	if (MapViewer::getTransCoordinates(this, m, t, true)) {
		m->ismodified = true;
		redrawMap();
		updateTransList();
		updateButtons();
	}
 	return 1;
}

long MapWindow::onCmdTransDelete(FXObject *, FXSelector, void *)
{
	Map *m = episode->maps[currentmap];
	Transporter *t = m->transporters[trans_list->getCurrentItem()];
	if (FXMessageBox::question(this, MBOX_YES_NO, "FX-Repton Editor",
 	    "Are you sure to delete the transporter found at %d, %d?", t->orig_x+1, t->orig_y+1) == MBOX_CLICKED_YES) {
		m->transporters.erase(m->transporters.begin() + trans_list->getCurrentItem());
		if (trans_list->getCurrentItem() == trans_list->getNumItems() - 1)
			trans_list->setCurrentItem(trans_list->getNumItems() - 2);
		else
			trans_list->setCurrentItem(trans_list->getCurrentItem());
		m->xy(t->orig_x, t->orig_y) = Sprite_Empty;
		m->ismodified = true;
  		delete t;
		redrawMap();
		updateTransList();
		updateButtons();
 	}
	return 1;
}

long MapWindow::onCmdInfoChange(FXObject *, FXSelector, void *)
{
	Map *m = episode->maps[currentmap];
	m->readonly = (readonly_check->getCheck() == TRUE);
	m->password = pwd_edit->getText().trim();
	m->comments = cmt_edit->getText().trim();
	m->time_limit = FXIntVal(time_edit->getText());
	m->ismodified = true;
	updateMapsList();
	updateButtons();
	return 1;
}

void MapWindow::updateMapsList()
{
	maps_list->clearItems();
	int n = 1;
  	typedef vector<Map *>::const_iterator I;
	for (I i = episode->maps.begin(); i != episode->maps.end(); ++i) {
		Map *m = *i;
		FXString s = FXStringFormat("%d: %s", n, m->password.text());
		maps_list->appendItem(s);
 		++n;
	}
	maps_list->setCurrentItem(currentmap);
}

void MapWindow::updateTransList()
{
	Map *m = episode->maps[currentmap];
	trans_list->clearItems();
  	typedef vector<Transporter *>::const_iterator J;
	for (J i = m->transporters.begin(); i != m->transporters.end(); ++i) {
		Transporter *t = *i;
		FXString s = FXStringFormat("%d, %d --> %d, %d", t->orig_x+1, t->orig_y+1, t->dest_x+1, t->dest_y+1);
		trans_list->appendItem(s);
 	}
}

void MapWindow::selectMap(int no)
{
	currentmap = no;
	updateMapsList();
	updateTransList();
	Map *m = episode->maps[currentmap];
	mapview->setMap(m, sprites);
	mapsize_label->setText(FXStringFormat("Map size: %dx%d", m->width, m->height));
	updateButtons();

	readonly_check->setCheck(m->readonly ? TRUE : FALSE);
	pwd_edit->setText(m->password);
	cmt_edit->setText(m->comments);
	time_edit->setText(FXStringVal(m->time_limit));
}

void MapWindow::spritesChanged()
{
	redrawTools();
	redrawMap();
}

